/**************************************************************************************************

Copyright (c) 2015   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_galactic_center.h"
#include  "mtk_orbital_element.h"
#include  "mtk_our_galaxy.h"
#include  "mtk_draw_common.h"
#include  "mtk_object.h"
#include  "mtk_objects.h"
#include  "mtk_stars.h"
#include  "mtk_stars_2.h"
#include  "mtk_label.h"
#include  "mtk_target.h"
#include  "mtk_switches.h"
#include  "mtk_mode.h"
#include  "mtk_colors.h"
#include  "mtk_solar_system.h"
#include  "mtk_calc_alpha.h"
#include  "mtk_dialog_box.h"
#include  "mtk_our_galaxy_rt.h"
#include  "mtk_OpenGL.h"
#include  "../common/astro/astro_const.h"
#include  "../common/astro/astrometry.h"
#include  <vector>



//***
extern GLuint  texStar;
extern fast_pow  pow_for_stars;
extern float  StarImagePowIndex;
//***



static GLuint texGalCenter;
const char *fn_tex_GalCenter = "galactic_center_bg.png";
static double Lv0_ss;





//-----------
//  Texture
//-----------
void
load_galactic_center_texture()
{
	SetSplashText("͌nSwĩeNX`Ǎ...", "Loading textures of background of the galactic center...");
	texGalCenter = LoadTexture(fn_tex_GalCenter, TexShrinkRate);
}



//-----------------------
//  ͌nS̔wi`
//-----------------------
void
draw_galactic_center_background(const mtk_draw_struct& ds)
{
	if (!bGalCenterBG)  return;


	float af, Rmax;
	af = 1.0f;
	if (ds.tyMode == MODE_3D) {
		const float dist = norm(get_our_galaxy_position() - ds.camera_base.get_pos());

		const float dist1 = 2000;
		const float dist2 = 3000;
		if (dist > dist1) {
			float t = (dist - dist1) / (dist2 - dist1);
			af = 1 - t;
		}

		if (af <= 0.0) return;
		Rmax = get_Rmax();
	}

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDisable(GL_DEPTH_TEST);
	glBindTexture(GL_TEXTURE_2D, texGalCenter);


	af *= 0.7;

	glColor4f(1, 1, 1, af);
	glPushMatrix();
	go_to_galactic_coord();
	glScalef(Rmax, Rmax, Rmax);
	glRotatef(180, 0, 0, 1);
	glFrontFace(GL_CW);
	gl_draw_sphere(10, true);
	glFrontFace(GL_CCW);
	glPopMatrix();

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}




//--------
//  Data
//--------
struct orbit_elem_ss {
	const char *name;
	double a, e, i, Omg, omg, tP, T;
	bool bSelected;
	char SpcType;
	double Temp;
};


// ͒SubNz[̐̋Ovf
// Ref: Gillessen, Eisenhauer, Trippe, et al. 2009, ApJ, 692, 1075
const orbit_elem_ss S_stars[] = {
	{"S1",	0.508,	0.496,	120.82,	341.61,	115.3,	0.95,	132,	true,	'B',	25000},
	{"S2",	0.123,	0.880,	135.25,	225.39,	63.56,	2.32,	15.8,	true,	'B',	25000},
	{"S4",	0.298,	0.406,	77.83,	258.11,	316.4,	-25.6,	59.5,	false,	'B',	25000},
	{"S5",	0.250,	0.842,	143.7,	109.0,	236.3,	-16.4,	45.7,	false,	'B',	13000},
	{"S6",	0.436,	0.886,	86.44,	83.46,	129.5,	63.0,	105,	false,	'B',	0},
	{"S8",	0.411,	0.824,	74.01,	315.90,	345.2,	-16.2,	96.1,	true,	'B',	25000},
	{"S9",	0.293,	0.825,	81.00,	147.58,	225.2,	-12.2,	58,		false,	'B',	25000},
	{"S12",	0.308,	0.900,	31.61,	240.4,	308.8,	-4.37,	62.5,	true,	'B',	13000},
	{"S13",	0.297,	0.490,	25.5,	73.1,	248.2,	4.90,	59.2,	true,	'B',	13000},
	{"S14",	0.256,	0.963,	99.4,	227.74,	339.0,	0.07,	47.3,	true,	'B',	13000},
	{"S17",	0.311,	0.364,	96.44,	188.06,	319.45,	-8.0,	63.2,	false,	'L',	0},
	{"S18",	0.265,	0.759,	116.0,	215.2,	151.7,	-4.0,	50,		false,	'E',	0},
	{"S19",	0.798,	0.844,	73.58,	342.9,	153.3,	5.1,	260,	false,	'K',	4000},
	{"S21",	0.213,	0.784,	54.8,	252.7,	182.6,	28.1,	35.8,	false,	'L',	0},
	{"S24",	1.060,	0.933,	106.3,	4.2,	291.5,	24.9,	398,	false,	'L',	0},
	{"S27",	0.454,	0.952,	92.91,	191.90,	308.2,	59.7,	112,	false,	'K',	4000},
	{"S29",	0.397,	0.916,	122,	157.2,	343.3,	21,		91,		false,	'E',	0},
	{"S31",	0.298,	0.934,	153.8,	103.0,	314.0,	13.8,	59.4,	false,	'B',	13000},
	{"S33",	0.410,	0.731,	42.9,	82.9,	328.1,	-32.1,	96,		false,	'E',	0},
	{"S38",	0.139,	0.802,	166.0,	286.0,	203,	3.0,	18.9,	false,	'L',	0},
	{"S66",	1.210,	0.178,	135.4,	96.8,	106,	-218,	486,	false,	'O',	34000},
	{"S67",	1.095,	0.368,	139.9,	106.0,	215.2,	-305,	419,	false,	'O',	34000},
	{"S71",	1.061,	0.844,	76.3,	34.6,	331.4,	-354,	399,	false,	'E',	0},
	{"S83",	2.785,	0.657,	123.8,	73.6,	197.2,	61,		1700,	false,	'O',	34000},
	{"S87",	1.260,	0.423,	142.7,	109.9,	41.5,	-353,	516,	false,	'O',	0},
	{"S96",	1.545,	0.131,	126.8,	115.78,	231.0,	-376,	701,	false,	'O',	0},
	{"S97",	2.186,	0.302,	114.6,	107.72,	38,		175,	1180,	false,	'O',	0}
};

const int Ns = sizeof(S_stars) / sizeof(S_stars[0]);




struct S_star {
	const char *name;
	orbital_elements oe;
	char	SpcType;
	bool	bSelected;
	int		TempIdx;
};

static std::vector<S_star>  vS_stars;



static matrix3d M_C2E;
static float L00;



//--------
//  Init
//--------
void
init_S_stars()
{
	const double arcsec_to_pc = 1.0/(60.0*60.0)*M_PI/180.0 * R_gal;

	const star_picture& stp = get_star_picture();


	// VʍW物Wnւ̕ϊs߂
	vector3d X, Y, Z;
	Z = normalize(-vGC);
	X = vector3d(0, 0, 1);	// V̖k
	Y = normalize(cross_product(Z, X));
	X = normalize(cross_product(Y, Z));

	const matrix3d M_C2R(X, Y, Z);
	M_C2E = M_R2E * M_C2R;


	vS_stars.clear();
	for (int i = 0; i < Ns; i++) {
		const orbit_elem_ss &oe0 = S_stars[i];
		S_star ss;


		// O
		ss.name = oe0.name;

		// Ovf̏
		double epoch, a, M0, T, n;

		epoch = oe0.tP * 365.25 + J2000;	// ߓ_ʉߎƂ
		M0 = 0.0;

		a = arcsec_to_pc * oe0.a;
		T = oe0.T * 365.25;	// PʂƂ
		n = 360.0 / T;		// xŕ\ω^

		ss.oe.set_parameters(epoch, a, oe0.e, M0, n, oe0.Omg, oe0.i, oe0.omg);


		// I
		ss.bSelected = oe0.bSelected;

		// XyNg^
		ss.SpcType = oe0.SpcType;

		// ̐FieNX`ԍj
		double Temp;
		if (oe0.Temp > 0) {
			Temp = oe0.Temp;
		}
		else{

			switch (ss.SpcType) {
			case 'O':	Temp = 34000;	break;
			case 'B':	Temp = 20000;	break;
			case 'K':	Temp = 4000;	break;
			case 'E':	Temp = 15000;	break;	// Early type
			case 'L':	Temp = 4400;	break;	// Late type
			default:	Temp = 2000;	break;
			}
		}
		ss.TempIdx = stp.get_TempIdx(Temp);


		//--  IuWFNgo^
		int idx = vS_stars.size();
		char key[256];
		sprintf(key, "S_star_%d", oe0.name);

		int obj_ID;
		obj_ID = register_object(mtk_object(OBJ_TYPE_S_STAR, idx, 0.0f, true), key);



		//--  xo^
		sprintf(key, "S_star_Label%d", oe0.name);

		//*******
		float R = 0;
		float thc = 8 * D2R;
		//*******

		add_label(oe0.name, oe0.name, obj_ID, MTK_COLOR_NAME_OBJECT, 1.0f, get_object_name_size(), R, thc, true, key);
		
		vS_stars.push_back(ss);


		//--  Oo^
		space_curve sc(mtk_object(OBJ_TYPE_S_STAR, i, 0.0f, false), get_orbit_width(), MTK_COLOR_ORBIT_PLANET);
		register_space_curve(sc);
	}


	// 邳̊
	L00 = pow(80.0, 1.0 / StarImagePowIndex);
}



bool
is_S_star_selected(int idx)
{
	return vS_stars[idx].bSelected;
}



//---------
//  Orbit
//---------
void
make_S_stars_orbit_curve(space_curve& sc, int div_N)
{
	const orbital_elements&  oe = vS_stars[sc.get_related_object().get_index()].oe;

	sc.clear();
	for (int j = 0; j<div_N; j++) {
		double phi = 2.0 * M_PI * double(j) / div_N;
		sc.add_point( M_C2E * oe.calc_orbit_pos(phi) + get_our_galaxy_position() );
	}
}




//------------
//  Position
//------------
void
update_S_stars_position(const double& t)
{
	for (std::vector<S_star>::iterator it = vS_stars.begin(); it != vS_stars.end(); it++) {
		it->oe.update_pos_and_velocity(t);
	}
}


vector3d
get_S_star_position(int idx)
{
	return M_C2E * vS_stars[idx].oe.get_pos() + get_our_galaxy_position();
}




//--------------
//  Visibility
//--------------
const double star_lc1 = 0.01 * _1AU_pc;
const double star_lc2 = 0.03 * _1AU_pc;


bool
is_S_star_label_visible(const mtk_draw_struct& ds, int idx)
{
	if (!(swStar & STAR_DISP))  return false;
	if (!(swStar & STAR_NAME))  return false;
	if ((swStar & STAR_NAME_SELECT) && !vS_stars[idx].bSelected)  return false;


	const double lc = norm(ds.camera_base.get_pos() - get_our_galaxy_position());
	
	//if (lc > 1.0)  return false;
	//if (lc < star_lc1)  return false;

	return  true;
}


float
get_S_star_label_alpha(const mtk_draw_struct& ds, int idx)
{
	const vector3d  cpos = ds.camera.get_pos();
	float  Dth = 7000.0f * _1AU_pc / norm(get_our_galaxy_position() - cpos);

	return  calc_alpha_by_apparent_diameter(Dth);
}




//---------
//  Alpha
//---------
float
get_S_star_alpha(const mtk_draw_struct& ds, int idx)
{
	bool bTgt = (get_target().get_type() == OBJ_TYPE_S_STAR && get_target().get_index() == idx);

	// ͌nS̐̈ʒu
	const vector3d  pos = M_C2E * vS_stars[idx].oe.get_pos();

	const float R1 = 0.01;
	const float R2 = 0.04;
	const float R3 = 1;
	const float R4 = 4;
	float R = norm(pos) / ds.scale;
	if (!bTgt && R > R4)  return  0.0;
	if (R < R1)  return  0.0;


	float af = 1.0;
	float sz = 1.5f / (ds.rad_per_pixel / 0.001);
	if (sz < 1) {
		af *= sz;
		sz = 1;
	}

	if (!bTgt && R > R3)	{ af *= 1 - (R - R3) / (R4 - R3); }
	if (R < R2)				{ af *= (R - R1) / (R2 - R1); }


	if (af < 0.05)  return 0.0;
	return  af;
}



float
get_S_star_object_orbit_alpha(const mtk_draw_struct& ds, int idx)
{
	const vector3d  cpos = ds.camera.get_pos();
	float  Dth = 20000.0f * _1AU_pc / norm(get_our_galaxy_position() - cpos);

	return  calc_alpha_by_apparent_diameter(Dth);
}


//---------
//  Scale
//---------
double
get_scale_S_star(int idx)
{
	return  200.0 * _1AU_pc;
}




//
//
//
void
prepare_draw_S_stars_2(const mtk_draw_struct &ds)
{
	float L0 = L00;

	switch (get_display_mode()) {
	case DISPMODE_SIDE_BY_SIDE:
	case DISPMODE_TOP_AND_BOTTOM:
		L0 *= 2.5;
		break;
	case DISPMODE_DOME_MASTER:
		L0 *= 4.0;
		break;
	default:
		break;
	}

	
	Lv0_ss = L0 / get_L_ref();
}




//----------------
//  Prepare draw
//----------------
double
get_star_radius_ss(const mtk_draw_struct& ds, const vector3d& rpos, const float& L)
{
	const float R1 = 1.0e-7;	// P 1pc ł̃C[Wa


	//  Radius
	double dist2 = norm2(rpos);
	float dist = sqrt(dist2);

	return pow_for_stars.eval(float(Lv0_ss*L / dist2)) * R1 * dist;
}


bool
prepare_draw_S_star(const mtk_draw_struct& ds, int idx, double* pRb, double* pRh)
{
	//---  Visible
	bool bFlg;
	bFlg = (swStar & STAR_DISP);

	if (!bFlg)  return false;
	

	//---  Radius
	vector3d rpos;
	double L;

	rpos = get_S_star_position(idx) - ds.camera_base.get_pos();
	L = Lv0_ss;

	*pRh = *pRb = get_star_radius_ss(ds, rpos, L);


	return  true;
}




//--------
//  Draw
//--------
void
draw_a_S_star(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl, const double& Rb, float alp)
{
	double R2;
	R2 = norm2(rpos);
	if (R2 < 1e-100)  return;


	const float sz = 2.0 * scl * Rb;



	if (true) {
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glDepthMask(GL_FALSE);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glBindTexture(GL_TEXTURE_2D, texStar);
		glColor3f(1, 1, 1);

		draw_a_star_TempIdx_2(ds, rpos, sz, vS_stars[idx].TempIdx, 1);
	}
	else {
		glPointSize(sz);
		glBegin(GL_POINTS);
		glVertex3dv(rpos.get_pointer());
		glEnd();
	}
}




//--------
//  Info
//--------
void
make_object_info_S_star(const mtk_draw_struct& ds, int idx, std::vector<std::string>& text)
{
	const S_star&  ss = vS_stars[idx];
	bool bEnglish = (ds.lang == LANG_ENG);
	char str[256];


	// O
	text.push_back(ss.name);

	// 
	if (bEnglish) {
		text.push_back("Type: Star");

		switch (ss.SpcType) {
		case 'E':
			text.push_back("Spactral type: Early type");
			break;
		case 'L':
			text.push_back("Spactral type: Late type");
			break;
		case 'K':
			text.push_back("Spactral type: K-M");
			break;
		default:
			sprintf(str, "Spactral type: %c", ss.SpcType);
			text.push_back(str);
			break;
		}
	}
	else {
		text.push_back("ށFP");

		switch(ss.SpcType) {
		case 'E':
			text.push_back("XyNg^F^");
			break;
		case 'L':
			text.push_back("XyNg^Fӊ^");
			break;
		case 'K':
			text.push_back("XyNg^FK-M");
			break;
		default:
			sprintf(str, "XyNg^F%c", ss.SpcType);
			text.push_back(str);
			break;
		}
	}
}
